/********************************************************************
 *                   NASA CONFIDENTIAL INFORMATION                  *
 *               (c)1997-2001 Nasa Ames Research Center			    *
 *                        All Rights Reserved                       *
 *                                                                  *
 *  This program contains confidential and proprietary information  *
 *  of NASA Ames Research Center, any reproduction, disclosure,     *
 *  or use in whole or in part is expressly prohibited, except as   *
 *  may be specifically authorized by prior written agreement.      *
 *                                                                  *
 ********************************************************************/


/*
ATM Model by Alessandro Acquisti acquisti@riacs.edu


SECTION 4.9 VERSION


THE STORY:
This simulation models the simple life of laborious Berkeley students: studying, eating, and getting just enough cash to afford some staple. 
Students spend most of their time studying, but get hungrier as the time goes by. When they are particularly hungry (the level varying depending on the student), they decide to move to one of the two restaurants in town. Students choose the restaurant according to how much money they are carrying. If a student does not have enough money for the cheapest restaurant, she will decide to pass first by the ATM of the bank where she has her account.  

When she arrives at the ATM, the student inserts her bank card and attempt to remember the PIN associated to her account. The ATM gives students 3 attempts to digit the correct PIN, before refusing the card altogether. The ATM communicates with the central bank computer to verify the correctness of the information provided by the student. If the bank computer communicates to the ATM that the PIN is correct and that the student has enough balance in her account, the ATM will dispense the cash and
will print an invoice with the account number and the remaining balance.

Students need to have enough balance in their  accounts to take out
cash: if they attempt to take out more money than they have, the bank computer
will notify the students (through the ATM) of the remaining amount of dollars in the
account. The student will modify her habits accordingly, and take out just exactly the
remaining dollars. 


THE CAST (AGENTS and OBJECTS):
Students: Kim, Alex
Bank computers: Bank of America, Wells Fargo
Restaurants: Blakes, Raleighs
Studying Places: South Hall, Spraul Hall
Clock: the Campanile
ATMs: one ATM for each bank


NOTES:
- attributes and relations scopes are all set to public. private and protected scopes are admitted but not yet implemented in the current version of the language
- the simulation halts at the end of a Berkeley's student hard work day: 20 hours (sleep is not considered!)
- except in the FINAL version/section, after taking money from the ATMs, students pass by the Hall before getting to the Restaurant
- objects are not given wait activities since detactables/facts do the job (see Tutorial text)


SUGGESTED CHANGES FOR FURTHER MODELING:
- start using private/public/protected scope (see above)
- change ATM and BANK to agents


FURTHER COMPLICATIONS:
- what if each bank has more than one ATM? can you make the agent choose the closest one?
- can  you model the objects (say, Banks and ATMs) as agents? what will change? do you notice changes in the behavior related to detectables and communications?
- what if more students go to the same ATM at the same time? can you make the ATM handle just one order at a time, but the Bank central computer multiple orders coming from several ATMs?
- can you make the agents randomly meet and interact (e.g., say hello each to the other)?
- what if you want the ATM to remember the number of the cards if has refused?


VERSION HISTORY
//VERSION 0.1, February 2001
//VERSION 0.9, April 2001
//VERSION 0.91 April 30, 2001
//VERSION 0.99 August 8, 2001
//VERSION 0.999 December 8, 2001

*/


package projects.atm;

class Atm {

	display: "Atm";
  	cost: 0.0;
 	resource: true;

	attributes:
		public int currentAccountCode;
		public int currentAccountPin;
		public boolean sessionIsOn;
		public boolean checkedAccountCode;
		public boolean checkedAccountPin;
		public boolean pinChecked;
		public boolean askedVerification;
		public boolean cashCanBeDispensed;


	relations:
		public Bank ownedbyBank;
		//ATTRIBUTES AND RELATIONS SCOPES ARE ALL SET TO PUBLIC. PRIVATE AND PROTECTED SCOPES ARE ADMITTED BUT NOT YET IMPLEMENTED IN THE CURRENT VERSION OF THE LANGUAGE

	initial_facts:
		(current.sessionIsOn = false);
		(current.checkedAccountCode = false);
		(current.checkedAccountPin = false);
		(current.pinChecked = false);
		(current.askedVerification = false);
		(current.cashCanBeDispensed != false);


	//NOTE: IN SECTION 4.9, YOU ARE NOT YET REQUESTED TO EXPLICITLY MODEL THE INTERACTION BETWEEN STUDENT, ATM AND BANK
	//THEREFORE, THE FOLLOWING ACTIVITIES AND WORKFRAMES ARE COMMENTED OUT AND OFFERED SIMPLY AS ROUGH, DRAFTY EXAMPLES OF HOW YOU COULD MODEL THE ATM AND THE BANK IN THE FOLLOWING SECTIONS

	activities:
/*
		primitive_activity getAccount() {
					max_duration: 25;
				}

		primitive_activity getPin() {
					max_duration: 35;
				}

		primitive_activity comparePins() {
					max_duration: 45;
				}

		primitive_activity matchPin_No() {
					max_duration: 25;
				}


		primitive_activity matchPin_Yes() {
					max_duration: 25;
				}

		primitive_activity returnCardandMoney() {
					max_duration: 25;
				}

		primitive_activity returnCardwithNoMoney() {
					max_duration: 25;
				}

		communicate replyStudent(Student std) {

					max_duration: 20;
					with: std;
					about:
		
						send(current.cashCanBeDispensed = anyvalue);
					
					when: end;
		}
		
		communicate askBankVerification_1(Bank ba3) {
					max_duration: 20;
					with: ba3;
					about:
					//	send(bkc.code = c);
						send(current.currentAccountCode = c),
						send(current.currentAccountPin = p);
					
					when: end;
		}

		communicate askBankVerification_2(Bank ba3, BankCard bkc) {
					max_duration: 20;
					with: ba3;
					about:

						send(current.currentAccountPin = p),
						send(ba3.sessionIsOn = true);
					
					when: end;
		}
*/

	workframes:

/*		
		workframe wf_getAccount {

					repeat: true;

					variables:

						forone(BankCard) bkc; 

					
					when(knownval(current contains bkc) and

						knownval(current.sessionIsOn = false) and
						knownval(current.checkedAccountCode = false))
						
						
					do {
						getAccount();
						conclude((current.currentAccountCode = bkc.code ), bc:100, fc:100);
						conclude((current.sessionIsOn = true), bc:100, fc:100);
						conclude((current.checkedAccountCode = true), bc:100, fc:100);
						
					}
		}

		workframe wf_getPin {

					repeat: true;

					variables:
						forone(Account) bka;   

						forone(BankCard) bkc; 


					
					when(knownval(current contains bkc) and
						known(bka.pin) and

						knownval(current.checkedAccountPin = false) and
						knownval(current.checkedAccountCode = true))
						
						
					do {
						getPin();
						conclude((current.currentAccountPin = bka.pin ), bc:100, fc:100);
						conclude((current.checkedAccountPin = true), bc:100, fc:100);
						
					}
		}
		

		workframe wf_askBankVerification {

					repeat: true;

					variables:
						forone(Account) bka;   

						forone(BankCard) bkc; 

						forone(Bank) ba3;

					
					when(knownval(current contains bkc) and
						knownval(current.checkedAccountPin = true) and
						knownval(current.checkedAccountCode = true) and
						knownval(current ownedbyBank ba3) and
						knownval(current.askedVerification = false))						
						
						
					do {
						conclude((ba3.sessionIsOn = true), bc:100, fc:100);
						askBankVerification_1(ba3);

						conclude((current.askedVerification = true), bc:100, fc:100);
											
					}
		}


		workframe wf_matchPin_yes {

					repeat: true;

					variables:
						forone(Bank) ba3;
						forone(BankCard) bkc;
				
											
					when(knownval(current ownedbyBank ba3) and
						knownval(current contains bkc) and

						knownval(ba3.authorizeCash = true) and
						knownval(current.pinChecked = false))
						
						
					do {
						matchPin_Yes();
						conclude((current.cashCanBeDispensed = true), bc:100);
						conclude((current.pinChecked = true), bc:100, fc:100);
										
					}
		}

		workframe wf_matchPin_no {

					repeat: true;

					variables:
						forone(Bank) ba3;
						forone(BankCard) bkc;
				
											
					when(knownval(current ownedbyBank ba3) and
						knownval(current contains bkc) and
						knownval(ba3.authorizeCash = false) and
						knownval(current.pinChecked = false))
						
						
					do {
						matchPin_No();
						conclude((current.cashCanBeDispensed = false), bc:100);
						conclude((current.pinChecked = true), bc:100, fc:100);
			
											
					}
		}
		
		workframe wf_returnCardandMoney {


					repeat: true;

					variables:
						forone(BankCard) bkc; 
						forone(Student) std;
					
					when(knownval(current contains bkc) and
						knownval(current.sessionIsOn = true) and
						knownval(current.checkedAccountPin = true) and
						knownval(std hasBankCard bkc) and
						knownval(current.checkedAccountCode = true) and
						knownval(current.pinChecked = true) and
						knownval(current.cashCanBeDispensed = true))
					
				

					do {
						replyStudent(std);
						conclude((current.sessionIsOn = false ), bc:100, fc:100);
						conclude((current contains bkc is false), bc:100, fc:100);
						conclude((std contains bkc), bc:100, fc:100);
						conclude((current.checkedAccountCode = false ), bc:100, fc:100);
						conclude((current.checkedAccountPin = false ), bc:100, fc:100);
						conclude((current.pinChecked = false), bc:100, fc:100);
						conclude((current.askedVerification = false), bc:100, fc:100);
					}
		}
		
				
		workframe wf_returnCardwithNoMoney {

					repeat: true;

					variables:
						forone(BankCard) bkc; 
						forone(Student) std;
					
					when(knownval(current contains bkc) and
						knownval(current.sessionIsOn = true) and
						knownval(current.checkedAccountPin = true) and
						knownval(std hasBankCard bkc) and
						knownval(current.checkedAccountCode = true) and
						knownval(current.pinChecked = true) and
						knownval(current.cashCanBeDispensed = false))
					
					

					do {
						replyStudent(std);
						conclude((current.sessionIsOn = false ), bc:100, fc:100);
						conclude((current contains bkc is false), bc:100, fc:100);
						conclude((std contains bkc), bc:100, fc:100);
						conclude((current.checkedAccountCode = false ), bc:100, fc:100);
						conclude((current.checkedAccountPin = false ), bc:100, fc:100);

						conclude((current.pinChecked = false), bc:100, fc:100);
						conclude((current.askedVerification = false), bc:100, fc:100);
						
					}
		}

*/
}

