/********************************************************************
 *                   NASA CONFIDENTIAL INFORMATION                  *
 *               (c)1997-2001 Nasa Ames Research Center			    *
 *                        All Rights Reserved                       *
 *                                                                  *
 *  This program contains confidential and proprietary information  *
 *  of NASA Ames Research Center, any reproduction, disclosure,     *
 *  or use in whole or in part is expressly prohibited, except as   *
 *  may be specifically authorized by prior written agreement.      *
 *                                                                  *
 ********************************************************************/


/*
ATM Model by Alessandro Acquisti acquisti@riacs.edu


SECTION 4.9 VERSION


THE STORY:
This simulation models the simple life of laborious Berkeley students: studying, eating, and getting just enough cash to afford some staple. 
Students spend most of their time studying, but get hungrier as the time goes by. When they are particularly hungry (the level varying depending on the student), they decide to move to one of the two restaurants in town. Students choose the restaurant according to how much money they are carrying. If a student does not have enough money for the cheapest restaurant, she will decide to pass first by the ATM of the bank where she has her account.  

When she arrives at the ATM, the student inserts her bank card and attempt to remember the PIN associated to her account. The ATM gives students 3 attempts to digit the correct PIN, before refusing the card altogether. The ATM communicates with the central bank computer to verify the correctness of the information provided by the student. If the bank computer communicates to the ATM that the PIN is correct and that the student has enough balance in her account, the ATM will dispense the cash and
will print an invoice with the account number and the remaining balance.

Students need to have enough balance in their  accounts to take out
cash: if they attempt to take out more money than they have, the bank computer
will notify the students (through the ATM) of the remaining amount of dollars in the
account. The student will modify her habits accordingly, and take out just exactly the
remaining dollars. 


THE CAST (AGENTS and OBJECTS):
Students: Kim, Alex
Bank computers: Bank of America, Wells Fargo
Restaurants: Blakes, Raleighs
Studying Places: South Hall, Spraul Hall
Clock: the Campanile
ATMs: one ATM for each bank


NOTES:
- attributes and relations scopes are all set to public. private and protected scopes are admitted but not yet implemented in the current version of the language
- the simulation halts at the end of a Berkeley's student hard work day: 20 hours (sleep is not considered!)
- except in the FINAL version/section, after taking money from the ATMs, students pass by the Hall before getting to the Restaurant
- objects are not given wait activities since detactables/facts do the job (see Tutorial text)


SUGGESTED CHANGES FOR FURTHER MODELING:
- start using private/public/protected scope (see above)
- change ATM and BANK to agents


FURTHER COMPLICATIONS:
- what if each bank has more than one ATM? can you make the agent choose the closest one?
- can  you model the objects (say, Banks and ATMs) as agents? what will change? do you notice changes in the behavior related to detectables and communications?
- what if more students go to the same ATM at the same time? can you make the ATM handle just one order at a time, but the Bank central computer multiple orders coming from several ATMs?
- can you make the agents randomly meet and interact (e.g., say hello each to the other)?
- what if you want the ATM to remember the number of the cards if has refused?


VERSION HISTORY
//VERSION 0.1, February 2001
//VERSION 0.9, April 2001
//VERSION 0.91 April 30, 2001
//VERSION 0.99 August 8, 2001
//VERSION 0.999 December 8, 2001
*/

package projects.atm;

group Student {

	attributes:

		public boolean male;
		public double howHungry;
		public double preferredCashOut;
		public int perceivedtime;
		public boolean readyToLeaveRestaurant;
		public Bank chosenBank;
		public boolean needCash;
		public boolean receivedCash;
		public boolean pinCommunicated;
		public boolean readyToLeaveAtm;
		public symbol hungriness;

		
	relations:
		public Account hasAccount;
		public Cash hasCash;
		public BankCard hasBankCard;

		//ATTRIBUTES AND RELATIONS SCOPES ARE ALL SET TO PUBLIC. PRIVATE AND PROTECTED SCOPES ARE ADMITTED BUT NOT YET IMPLEMENTED IN THE CURRENT VERSION OF THE LANGUAGE
		
	initial_beliefs:

		(Blakes_Diner.location = Telegraph_Av_2134);
		(Raleigh_Diner.location = Telegraph_Av_2405);
		(Boa_Atm.location = Telegraph_Av_113);
		(current.perceivedtime = 1);
		(Campanile_Clock.time  = 1);
		(current.readyToLeaveRestaurant = false);

	initial_facts:
				
	activities:

			composite_activity study() {
				activities:
					primitive_activity reading ( ) {
						max_duration: 1500;
					}
					
				workframes:
					workframe wf_readingWhileStudying {
						do {
							reading();
						}
					}
										
				thoughtframes:
					thoughtframe tf_needCashToEat {
						variables:
							forone(Cash) cs;		
						when (knownval(cs.amount < current.preferredCashOut))
						do {
							conclude((current.needCash = true));
						}
					}
					
					thoughtframe tf_enoughCashToEat {
						variables:
							forone(Cash) cs;		
						when (knownval(cs.amount >= current.preferredCashOut))
						do {
							conclude((current.needCash = false));
						}
					}
					
					thoughtframe tf_HowMuchMoneyToGet_HungryEQhigh {
						when(knownval(current.needCash = true) and
						 		 knownval(current.hungriness = high))
						do {
							conclude((current.preferredCashOut = 15));
						}
					}
					
					thoughtframe tf_HowMuchMoneyToGet_HungryEQlow {
						when(knownval(current.needCash = true) and
						 		 knownval(current.hungriness = low))
						do {
							conclude((current.preferredCashOut = 5));
						}
					}		
					
					thoughtframe tf_HowMuchMoneyToGet_HungryEQmedium {
						when(knownval(current.needCash = true) and
								knownval(current.hungriness = medium))
						do {
							conclude((current.preferredCashOut = 10));
						}
					}	
						
					thoughtframe tf_Hungriness_High {
						variables:
							forone(double) n_howHungry;
						when(knownval(current.howHungry = n_howHungry) and
								 knownval(n_howHungry > 20))
						do {
							conclude((current.hungriness = high));
						}
					}
					thoughtframe tf_Hungriness_Low {
						variables:
							forone(double) n_howHungry;
						when(knownval(current.howHungry = n_howHungry) and
								 knownval(n_howHungry <= 10))
						do {
							conclude((current.hungriness = low));
						}
					}
					
					thoughtframe tf_Hungriness_Medium {
						variables:
							forone(double) n_howHungry;
						when(knownval(current.howHungry = n_howHungry) and
								 knownval(n_howHungry >= 10) and
								 knownval(n_howHungry <= 20))
						do {
							conclude((current.hungriness = medium));
						}		
					}				
			}	
		
			move moveToRestaurant() {
					

					location: Telegraph_Av_2405;
			}

			move moveToSouthHall() {
					
					location: SouthHall;
			}

			move moveToLocationForCash(Building loc) {
					location: loc;
			}

			
			move moveToLocation(Building loc) {
					location: loc;
			}
	
			
			primitive_activity eat() {
					max_duration: 400;

			}

		composite_activity useATM() {

			activities:
				
							  
				primitive_activity insertBankCard() {
					max_duration: 45;
				}
		
				
				communicate communicatePIN(Atm at3, Account bka) {
					max_duration: 20;
					with: at3;
					about:
						send(bka.pin = unknown);
			

					when: end;
				}
			 
				/* TO BE REINSERTED FOR FINAL SECTION
				primitive_activity waitAtmReply() {
					max_duration: 1000;
				}
				*/

				primitive_activity getCash() {
					max_duration: 50; 
				}

				/* TO BE REINSERTED FOR FINAL SECTION
				primitive_activity doNotGetCashAndCardAndInvoice() {
					 max_duration: 50;
				}
				*/
				
			workframes:
				
				
				workframe wf_moveToLocationForCash {
					repeat: true;
				
					variables:
						forone(Cash) cs; 
						forone(Atm) at;
						forone(Bank) bk;
						forone(Building) bd;
					
					when(knownval(current hasCash cs) and
						knownval(current.chosenBank = bk) and
						knownval(at ownedbyBank bk) and
						knownval(cs.amount < 10.00) and
						not(current.location = at.location) and
						knownval(at.location = bd ))
						
						
					do {
						moveToLocation(bd);

					}
				}  
					
				
				workframe wf_insertBankCard {
					
					repeat: true; 
					
					variables:
						forone(BankCard) bkc2; 
						forone(Atm) at2;
						forone(Bank) bk2;
						forone(Building) bd2;
					
					when(knownval(current hasBankCard bkc2) and
						knownval(current.chosenBank = bk2) and
						knownval(at2 ownedbyBank bk2) and
						knownval(current.receivedCash = false) and
						knownval(current.location = at2.location) and
						knownval(current.pinCommunicated = false) and
						knownval(current contains bkc2)) 
									
					
					do {
						insertBankCard();
						conclude((current contains bkc2 is false), bc:100, fc:100);	
						conclude((at2 contains bkc2), bc:100, fc:100);	
					}
				}

				workframe wf_communicatePIN {
					
					repeat: true;

					variables:
						forone(Account) bka;   
						forone(BankCard) bkc3; 
						forone(Atm) at3;
						forone(Bank) ba3;
						forone(Building) bd3;

					
					when(knownval(current hasBankCard bkc3) and
						not(current contains bkc3) and
						knownval(current hasAccount bka) and
						knownval(current.chosenBank = ba3) and
						knownval(at3 ownedbyBank ba3) and
						knownval(current.pinCommunicated = false) and
						knownval(current.location = at3.location) and
						knownval(at3 contains bkc3)) 
						
						
					do {
						communicatePIN(at3, bka);
						conclude((current.pinCommunicated = true), bc:100, fc:0);
						conclude((current contains bkc3), bc:100, fc:100);
						conclude((at3 contains bkc3 is false), bc:100, fc:100);

					}
				}

				/*TO BE RESTORED IN FINAL SECTION
				workframe wf_waitAtmReply {
				
					repeat: true; 

					variables:
						forone(BankCard) bkc4;         
						forone(Cash) cs4; 
						forone(Atm) at4;
						forone(Bank) bk4;
					
					detectables:
						detectable atmRepliesYes {
							when(whenever)
								detect((at4.cashCanBeDispensed = true), dc:100)
								then abort;
						}

						detectable atmRepliesNo {
							when(whenever)
								detect((at4.cashCanBeDispensed = false), dc:100)
								then abort;
						}
					
					when(knownval(current hasBankCard bkc4) and
						knownval(current hasCash cs4) and
						not(current contains bkc4) and
						knownval(current.chosenBank = bk4) and
						knownval(at4 ownedbyBank bk4) and
						knownval(current.pinCommunicated = true) and
						not(at4.cashCanBeDispensed = true) and
						not(at4.cashCanBeDispensed = false))

					do {
						waitAtmReply();
											
					}
				
				}
				/*
				
				  
				/* TO BE RESTORED IN FINAL SECTION
				workframe wf_doNotGetCashAndCardAndInvoice {
				
					repeat: true; 

					variables:
						forone(BankCard) bkc4;         
						forone(Cash) cs4; 
						forone(Atm) at4;
						forone(Bank) bk4;
						
					
					when(knownval(current.hasCard = false) and
						knownval(current hasBankCard bkc4) and
						knownval(current hasCash cs4) and
						knownval(current.chosenBank = bk4) and
						knownval(at4 ownedbyBank bk4) and
						knownval(current.pinCommunicated = true) and
						knownval(at4.cashCanBeDispensed = false))

					do {
						doNotGetCashAndCardAndInvoice();
						conclude((current contains bkc4), bc:100, fc:100);
						conclude((at4 contains bkc4 is false), bc:100, fc:100);
						conclude((current.hasCard = true), bc:100, fc:0);
						conclude((current.pinCommunicated = false), bc:100, fc:0);
						conclude((current.readyToLeaveAtm = true), bc:100, fc:0);
						conclude((at4.cashCanBeDispensed != true), bc:100, fc:0);
						
					}
				
				}

				*/
					
				
				workframe wf_getCash {
				
					repeat: true; 

					variables:
						forone(BankCard) bkc4;         
						forone(Cash) cs4; 
						forone(Atm) at4;
						forone(Bank) bk4;
						
					
					when(knownval(current hasBankCard bkc4) and
						knownval(current hasCash cs4) and
						knownval(current contains bkc4) and
						knownval(current.chosenBank = bk4) and
						knownval(at4 ownedbyBank bk4) and
						knownval(current.pinCommunicated = true))

					do {
						getCash();
						conclude((cs4.amount = cs4.amount + current.preferredCashOut), bc:100, fc:100); 
						conclude((current.pinCommunicated = false), bc:100, fc:0);
						conclude((current.receivedCash = true), bc:100, fc:0);
						conclude((current.readyToLeaveAtm = true), bc:100, fc:0);
					}
				
				}
				
			
				workframe wf_backToStudy {
					repeat: true;
				
					variables:
						forone(Atm) at5;

					when(knownval(current.readyToLeaveAtm = true) and
						knownval(current.location = at5.location))

						do {
							
							moveToLocation(SouthHall);
							conclude((current.needCash = false), bc:100, fc:100);
							conclude((current.readyToLeaveAtm = false), bc:100, fc:0);
							conclude((current.receivedCash = false), bc:100, fc:0);

				//NOTE: CURRENTLY WE MAKE THE STUDENTS GO BACK TO A UNIVERSITY HALL AFTER GETTING THE MONEY AND BEFORE SPENDING THE MONEY AT SOME RESTAURANT			
				//BECAUSE IT IS EASIER. BUT IN THE FINAL SECTION WE WILL MODEL THE MORE REALISTIC CASE WHERE THE AGENT GOES STRAIGHT TO THE RESTAURANT
						
						}

				}

			}





	workframes:


		/* THIS WORKFRAME IS MODIFIED (AND HENCE COMMENTED OUT) IN SECTION 4.7

		workframe wf_moveToRestaurant {
				
				repeat: true;
		
				
				variables:


				when
					(knownval(current.howHungry > 20.00) and
					knownval(current.location != Telegraph_Av_2405))
				do
					{moveToRestaurant();
					conclude((current.howHungry = current.howHungry - 5.00), bc:100, fc:0);
					conclude((Alex_Cash.amount = Alex_Cash.amount - Raleigh_Diner.foodcost), bc:100, fc:100);
					moveToSouthHall();
				}
		}
		*/


				workframe wf_moveToRestaurant {
					repeat: true;

					variables:
						forone(Cash) cs; 
					
					when(knownval(current.howHungry > 20.00) and
						knownval(current hasCash cs) and
						knownval(cs.amount > 9.00) and
						knownval(current.location != Telegraph_Av_2405))
				
						
					do {
						moveToLocation(Telegraph_Av_2405);	
					}
				}
		
				workframe wf_Eat {

					repeat: true;

				
						
					when(knownval(current.howHungry > 20.00) and
						knownval(current.location = Telegraph_Av_2405))
										
					do {
						
						eat();
						conclude((current.howHungry = current.howHungry - 5.00), bc:100, fc:0);
						conclude((Alex_Cash.amount = Alex_Cash.amount - Raleigh_Diner.foodcost), bc:100, fc:100);
						conclude((current.readyToLeaveRestaurant = true), bc:100, fc:0);
					}


				}
			
				
				workframe wf_backToStudy {
					repeat: true;

					when(knownval(current.readyToLeaveRestaurant = true) and
						knownval(current.location = Telegraph_Av_2405))

						do {
							moveToLocation(SouthHall);
							conclude((current.readyToLeaveRestaurant = false), bc:100, fc:0);

						} 

				}
		
		/* THIS CORRECTED AFTER SECTION 4.7
		workframe wf_moveToLocationForCash {
					repeat: true;
				
					variables:
					
					
					when(knownval(Alex_Cash.amount < 10.00)) 
						
					do {
						moveToLocationForCash(Telegraph_Av_113);
						//MoveToLocation(Bancroft_Av_77);
						//before, it was: MoveToLocation(bankofamerica etc. without this generalization: an easier case...)
						conclude((Alex_Cash.amount = Alex_Cash.amount + current.preferredCashOut), bc:100, fc:100); 
						moveToLocationForCash(SouthHall);
					}
				}
		*/

		/* THIS COMMENTED OUT AFTER SECTION 4.8
		workframe wf_moveToLocationForCash {
					repeat: true;
				
					variables:
					
					
					when(knownval(Alex_Cash.amount < 10.00)) 
						
					do {
						moveToLocation(Telegraph_Av_113);
						conclude((Alex_Cash.amount = Alex_Cash.amount + current.preferredCashOut), bc:100, fc:100); 
						moveToLocation(SouthHall);
					}
				}
		*/


		/* THIS COMMENTED OUT WHEN STARTING SECTION 4.9
		workframe wf_moveToLocationForCash {
					repeat: true;
				
					variables:
						forone(Cash) cs; 
						forone(Atm) at;
						forone(Bank) bk;
						forone(Building) bd;
						forone(Account) ac;
					
					when(knownval(current hasCash cs) and
						knownval(cs.amount < 10.00) and
						knownval(current hasAccount ac) and
						knownval(ac openedWithBank bk) and
						knownval(at ownedbyBank bk) and
						not(current.location = at.location) and
						knownval(at.location = bd ) and
						knownval(current.readyToLeaveRestaurant = false))

						
					do {
						moveToLocation(bd);
						conclude((cs.amount = cs.amount + current.preferredCashOut), bc:100, fc:100); 
						moveToLocation(SouthHall);

					}
				}
			*/

				workframe wf_useATM {
					
					repeat: true;
        
											
					variables:
						forone(Cash) cs; 
						forone(Atm) at;
						forone(Bank) bk;
						forone(Building) bd;
						forone(Account) ac;
					
					when(knownval(current.howHungry > 20.00) and
						knownval(current hasCash cs) and
						knownval(cs.amount < 10.00) and
						knownval(current hasAccount ac) and
						knownval(ac openedWithBank bk) and
						knownval(at ownedbyBank bk) and
						not(current.location = at.location) and
						knownval(at.location = bd ) and
						knownval(current.readyToLeaveRestaurant = false))


					do {
						useATM();
						
					}
				}


			workframe wf_study {
					repeat: true;

					variables:
						forone(Cash) cs; 
		
	     
					when(knownval(Campanile_Clock.time < 20) and
						knownval(current.howHungry < 21) and
						knownval(current.location = SouthHall))
						
						do { 
							
							study();
							
						}
			}



	thoughtframes:

			thoughtframe tf_feelHungry {
					repeat: true;

					when(knownval(Campanile_Clock.time > current.perceivedtime))

					do {
						conclude((current.perceivedtime = Campanile_Clock.time), bc: 100);
						conclude((current.howHungry = current.howHungry + 3.00), bc:100);
					}

				}


}



